#include                <unistd.h>
#include                <string.h>
#include                <stdlib.h>
#include                <dlfcn.h>
#include		"tap.h"
#include		"libFireBird.h" // tested with FireBirdLib 2012-04-23
#include    "TMSCommander.h" //v1.5
#include    "useful.c" //v1.6_debug, v1.7 changed
#include		"ResetCAM_TAPCOM.h" //v1.8

#define PGM_NAME	"ResetCAM" //v1.6_debug changed
#define VERSION			"V1.8"
#define D(x)			if (_debug) x; //v1.6_debug changed, v1.7 changed
#define LOG_FILE  "ResetCAM.log" //v1.6_debug

TAP_ID				(0x8E0Affff);
TAP_PROGRAM_NAME	(PGM_NAME" "VERSION); //v1.6_debug changed
TAP_AUTHOR_NAME		("srupprecht & xyzzyx");
TAP_DESCRIPTION		("Magically resets your CAM");
TAP_ETCINFO			(__DATE__);

#define TICKS_PER_SEC	(100)

enum {
	CAM_CURRENT = -2,
	CAM_BOTH = -1,
	CAM_ONE = 0,
	CAM_TWO = 1
};

static dword _svcStatus[2] = {0,0};
static int _state = 0, _svcType = -1, _svcNum = -1, _slot = 0, _timeout = 5,
		   _checksub = 0, _end = 0, _badsq = 25;
static dword _checktime = 0, _key1code = 0, _key0code = 0, _nextresettime = 0;
static void *pvr;
static char *_currentSelectedCam;
static void (*Cam_ForcedReset)(unsigned int);
static bool _debug = 0; //v1.7
static int  counter = 0; //v1.8 moved here

/* stolen from Overfly */
/* v1.7 deleted >>>
#define KEY_COUNT		29

static char *keyName[KEY_COUNT] =
{
	"Recall",	"Mute",     "UHF",       "Sleep",
    "AudioTrk",	"Exit",     "TvRadio",   "Pause",
	"Fav",      "Subt",     "TvSat",     "Green",
	"Yellow",   "Blue",     "Teletext",  "Rewind",
	"Play",     "Forward",  "Slow",      "Stop",
	"White",    "Red",      "Sat",       "Prev",
	"Next",     "NoUse",    "Option",    "0",
	"1"
};

static int keyCode[KEY_COUNT] =
{
	0x1000B,0x1000C,0x1000E,0x10010,
	0x10013,0x10017,0x10018,0x1001A,
	0x10020,0x10021,0x10022,0x10024,
	0x10025,0x10026,0x10037,0x10038,
	0x10039,0x1003A,0x1003B,0x1003C,
	0x1003E,0x1003F,0x10040,0x10041,
	0x10043,0x10044,0x1003E,RKEY_0,
	RKEY_1
};
// v1.7 deleted <<< */
  
bool DoReset(int slot); //1.8 
static dword TMSCommander_handler(dword param1); //v1.5
dword TAP_EventHandler(word event, dword param1, dword param2);
//static dword GetKeyCodeByName(char *buf); v1.7 deleted
int TAP_Main(void);


// v1.8 >>>
bool DoReset(int slot)
{
	dword timenow;

	timenow = TAP_GetTick();
	if (timenow >= _checktime)
	{
				
		if (slot == CAM_BOTH)
		{
			Cam_ForcedReset(0);
			D(TAP_DEBUG("Reset (Both, this was slot 0)! Waiting 15 seconds.\n")); //v1.6_debug
			Cam_ForcedReset(1);
			D(TAP_DEBUG("Reset (Both, this was slot 1)! Waiting 15 seconds.\n")); //v1.6_debug
		}
		else if (slot == CAM_CURRENT)
		{
			Cam_ForcedReset(*_currentSelectedCam);
			D(TAP_DEBUG("Reset (Selected %ld)! Waiting 15 seconds.\n", *_currentSelectedCam)); //v1.6_debug moved down
		}
		else
		{
			Cam_ForcedReset(slot);
			D(TAP_DEBUG("Reset (Slot %ld)! Waiting 15 seconds.\n", slot)); //v1.6_debug
		}
				
		_nextresettime = timenow + (TICKS_PER_SEC * 15);
		_state = (++counter > 2) ? 2 : 0;
		return TRUE; //v1.8
	}
	D(TAP_DEBUG("Reset request too early\n")); //v1.8
	return FALSE; //v1.8
}
// v1.8 <<<

// TMSCommander_handler //v1.5
static dword TMSCommander_handler(dword param1)
{
	switch (param1)
	{
		case TMSCMDR_Capabilities:
			return (dword)(TMSCMDR_CanBeStopped);

		case TMSCMDR_Stop:
		//Exit your TAP
		  D(TAP_DEBUG("Goodbye.\n"));
	  	dlclose(pvr);
	  	_end = 1;
	  	TAP_Exit();
			return TMSCMDR_OK;

		case TMSCMDR_Exiting:
		//TMSCommander is exiting - do you want to do anything?
			return TMSCMDR_OK;

		default:
		//unknown TMSCommander function
			return TMSCMDR_UnknownFunction;
	}
}

dword TAP_EventHandler(word event, dword param1, dword param2)
{
  (void) param2;

  if (event == EVT_TMSCommander) //v1.5
		return TMSCommander_handler(param1); //v1.5

	// v1.8 >>>
  if(event == EVT_TAPCOM)
  {
    TAPCOM_Channel  Channel;
    dword           ServiceID;
    dword 					CallerID;

    Channel = TAPCOM_GetChannel(param1, &CallerID, &ServiceID, NULL, NULL);
    if(ServiceID == (dword) TAPCOM_SERVICE_RESETCAM_CURRENT ||
    	 ServiceID == (dword) TAPCOM_SERVICE_RESETCAM_BOTH ||
    	 ServiceID == (dword) TAPCOM_SERVICE_RESETCAM_ONE ||
    	 ServiceID == (dword) TAPCOM_SERVICE_RESETCAM_TWO)
    {
    	D(TAP_DEBUG("Good TAPCOM request received from CallerID 0x%08X with ServiceID 0x%08X\n", CallerID, ServiceID)); 
      TAPCOM_Finish(Channel, TAP_GetTick()<_nextresettime?0:(DoReset(ServiceID)));
    }	 
    else
    {
    	D(TAP_DEBUG("Bad TAPCOM request received from CallerID 0x%08X with ServiceID 0x%08X\n", CallerID, ServiceID)); 
      TAPCOM_Reject(Channel);
    }	 
  }
  // v1.8 <<<

  if (_end)
	  return param1;
  
  if (event == EVT_IDLE)
  {
    if (((_svcStatus[CHANNEL_Main] & SVCSTATUS_MASK_Error) != 0) ||
		(_checksub &&
		 ((_svcStatus[CHANNEL_Sub ] & SVCSTATUS_MASK_Error) != 0)))
    {
      switch (_state)
      {
		case 0:
		{
			TYPE_TapChInfo chInfo;
			dword timenow;
			
			if (!_timeout)
				break;

			if (TAP_GetSignalQuality() < _badsq)
			{
				D(TAP_DEBUG("Bad signal quality\n"));
				break;
			}

			timenow = TAP_GetTick();
			if (_nextresettime > timenow)
				break;

			TAP_Channel_GetCurrent(&_svcType, &_svcNum);

			if (TAP_Channel_GetInfo(_svcType, _svcNum, &chInfo))
			{
				if ((chInfo.flag & 0x04) == 0)
				{
					_state = 2;
					break;
				}
			}
		
			D(TAP_DEBUG("Start counter\n"));
			
			_checktime = timenow + _timeout;
			_state  = 1;
			counter = 0;
		}
		break;

		case 1:
		{
			int type, num;

			TAP_Channel_GetCurrent(&type, &num);
			if (type != _svcType || num != _svcNum)
			{
				D(TAP_DEBUG("Channel changed before Reset\n"));
				_state = 0;
				_nextresettime = 0;
				break;
			}

			DoReset(_slot); //v1.8
		}
		break;

		case 2:
		{
			int type, num;

			TAP_Channel_GetCurrent(&type, &num);
			if (type != _svcType || num != _svcNum)
			{
				_state = 0;
				_nextresettime = 0;
			}
		}
		break;
      }
    }
    else
    {
		_state = 0;
    }
  }
  else if (event == EVT_SVCSTATUS)
  {
    _svcStatus[(param1 & SVCSTATUS_MASK_MainSub) >> 28] = param1;
	  D(if (_svcStatus[CHANNEL_Main] & SVCSTATUS_MASK_Error) TAP_DEBUG("svcStatus Err Main %x\n", _svcStatus[CHANNEL_Main] & SVCSTATUS_MASK_Error)); //v1.6_debug
  	D(if (_svcStatus[CHANNEL_Sub] & SVCSTATUS_MASK_Error) TAP_DEBUG("svcStatus Err Sub %x\n", _svcStatus[CHANNEL_Sub] & SVCSTATUS_MASK_Error)); //v1.6_debug
  }
  else if (event == EVT_KEY)
  {
	  dword state = 0, substate = 0, timenow;
	  int slot;

	  if (_key1code && param1 == _key1code) slot = 1;
	  else if (_key0code && param1 == _key0code) slot = 0;
	  else
	  {
		  return param1;
	  }

	  timenow = TAP_GetTick();
	  if (_nextresettime > timenow)
			return 0;

	  TAP_GetState(&state, &substate);
	  if (state == STATE_Normal && substate == SUBSTATE_Normal)
	  {
		TYPE_PlayInfo playInfo;

		if (TAP_Hdd_GetPlayInfo(&playInfo) &&
			playInfo.playMode == PLAYMODE_None)
		{
			_state = 0;
			param1 = 0;

			Cam_ForcedReset(slot);
			_nextresettime = timenow + (TICKS_PER_SEC * 15);
			D(TAP_DEBUG("Key reset (Slot %d)! Waiting 15 seconds.\n", slot)); //v1.6_debug moved down and changed
		}
	  }
  }
  else if (event == EVT_STOP)
  {
	  D(TAP_DEBUG("Goodbye.\n"));
	  dlclose(pvr);
	  _end = 1;
	  TAP_Exit();
  }  

  return param1;
}

/* v1.7 deleted >>>
static dword GetKeyCodeByName(char *buf)
{
	int i, len = strlen(buf);

	if (len)
	{
		for (i = 0; i < KEY_COUNT; i++)
		{
			if (!strncasecmp(keyName[i], buf, len))
			{
				return keyCode[i];
			}
		}
	}

	return 0;
}
// v1.7 deleted <<< */

int TAP_Main(void)
{
	// v1.6_debug >>>
	char buf[256];
	
  TAP_SPrint(buf, "/ProgramFiles/Settings/%s", PGM_NAME);
  TAP_Hdd_ChangeDir(buf);
  if (!TAP_Hdd_Exist(LOG_FILE))
    TAP_Hdd_Create(LOG_FILE, ATTR_NORMAL);
	// v1.6_debug <<<

  char *funcname = "Cam_ForcedReset";

	KeyTranslate(TRUE, &TAP_EventHandler); //v1.6

  if (INIOpenFile(PGM_NAME".ini", PGM_NAME) != INILOCATION_NotFound) //v1.6_debug changed
  {
	char buf[32];

	_debug = INIGetInt("Debug",0,0,1); //v1.7

	_badsq = INIGetInt("BadSQ",25,0,60);
	
	_slot = INIGetInt("Slot",0,-2,1);
	
	_timeout = INIGetInt("TimeOut",_timeout,0,300);
	if (_timeout == 1)
		_timeout = 2;

	if (INIGetString("ResetType",buf,"reset",sizeof(buf)))
	{
		if (!strncasecmp(buf, "init", 4))
			funcname = "Cam_Init";
	}

	if (INIGetString("CheckSub",buf,"no",sizeof(buf)))
	{
		if (!strncasecmp(buf, "yes", 3))
			_checksub = 1;
	}

	if (INIGetString("ResetKey0",buf,"",sizeof(buf)))
	{
		_key0code = GetKeyValueByText(buf); //v1.7 changed
	}

	if (INIGetString("ResetKey1",buf,"",sizeof(buf)))
	{
		_key1code = GetKeyValueByText(buf); //v1.7 changed
	}

	D(TAP_DEBUG("Slot %d TimeOut %d Key0 %x Key1 %x CheckSub %ld Func %s SQ %ld\n",
				_slot, _timeout, _key0code, _key1code, _checksub, funcname, _badsq));
  }

  INICloseFile();
  
  _timeout *= TICKS_PER_SEC;

  //Get access to the exported symbols of the main app
  pvr = dlopen(NULL, RTLD_GLOBAL | RTLD_LAZY);
  if(!pvr) return 0;

  //Get the address of the function "Cam_ForcedReset(unsigned int)"
  Cam_ForcedReset = dlsym(pvr, funcname);
  if (dlerror())
  {
	D(TAP_DEBUG("reset function not found!\n"));
    dlclose(pvr);
    return 0;
  }

  dword *svcStatusPvr;

  svcStatusPvr = dlsym(pvr, "_svcStatus");
  if (dlerror())
  {
     D(TAP_DEBUG("_svcStatus not found!\n"));
    return 0;
  }

  _currentSelectedCam = dlsym(pvr, "_currentSelectedCam");
  if (dlerror())
  {
     D(TAP_DEBUG("_currentSelectedCam not found!\n"));
    return 0;
  }
  
  _svcStatus[CHANNEL_Main] = (*svcStatusPvr << 16) & SVCSTATUS_MASK_Error;
  _svcStatus[CHANNEL_Sub ] =  *svcStatusPvr        & SVCSTATUS_MASK_Error;

  D(TAP_DEBUG("svcStatus Err Main %x\n", _svcStatus[CHANNEL_Main] & SVCSTATUS_MASK_Error));
  D(TAP_DEBUG("svcStatus Err Sub %x\n", _svcStatus[CHANNEL_Sub] & SVCSTATUS_MASK_Error));

  return 1;
}
